/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Properties;

import javax.jbi.JBIException;
import javax.jbi.component.InstallationContext;

import org.apache.commons.io.FileUtils;
import org.objectweb.petals.component.framework.DefaultBootstrap;

/**
 * The bootstrap class. Copy all the configuration files in the good
 * directories.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public class SoapBootstrap extends DefaultBootstrap {

    protected SoapBootstrapOperations bootstrapOperations;

    /*
     * non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.basic.AbstractBasicComponent#init(javax.jbi.component.InstallationContext)
     */
    @Override
    public void init(final InstallationContext installContext)
            throws JBIException {
        super.init(installContext);
        this.createWorkDirectories();
        this.copyConfigurationFiles();

        if (bootstrapOperations == null) {
            bootstrapOperations = new SoapBootstrapOperations(installContext);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.basic.AbstractBasicComponent#cleanUp()
     */
    @Override
    public void cleanUp() throws JBIException {
        super.cleanUp();
    }

    /**
     * Create the directories that will be used by axis
     *
     * @param installPath
     */
    protected void createWorkDirectories() {
        File modules = new File(installContext.getInstallRoot(), "modules");
        File services = new File(installContext.getInstallRoot(), "services");
        modules.mkdirs();
        services.mkdirs();
    }

    /**
     * Copy the needed configuration files to the Axis2 directories. Do not copy
     * them if the files are already present (the install phase has already been
     * done and we are now in the uninstall phase).
     *
     */
    protected void copyConfigurationFiles() throws JBIException {

        File installRootFile = new File(installContext.getInstallRoot());

        // copy .mar files to the modules directory
        File modules = new File(installRootFile, "modules");
        File[] metaInfFiles = new File(installRootFile, "META-INF").listFiles();

        for (File file : metaInfFiles) {

            // copy modules from META-INF
            if (file.getName().endsWith(".mar")) {
                File destFile = new File(modules, file.getName());
                if (!destFile.exists()) {
                    try {
                        FileUtils.copyFile(file, destFile);
                    } catch (IOException e) {
                        // do nothing
                    }
                }
            }

            // copy axis2.xml file
            if (file.getName().equals("axis2.xml")) {
                File destFile = new File(installContext.getInstallRoot(), file
                        .getName());
                if (!destFile.exists()) {
                    try {
                        FileUtils.copyFile(file, destFile);
                    } catch (IOException e) {
                        throw new JBIException(
                                "Can not copy the axis2.xml file");
                    }
                }
            }
        }
        
        File[] rootFiles = installRootFile.listFiles();

        for (File file : rootFiles) {

            // copy modules from root
            if (file.getName().endsWith(".mar")) {
                File destFile = new File(modules, file.getName());
                if (!destFile.exists()) {
                    try {
                        FileUtils.copyFile(file, destFile);
                    } catch (IOException e) {
                        // do nothing
                    }
                }
            }
        }
    }

    @Override
    protected void doComponentConfiguration(Properties componentProperties,
            List<String> methods) {

        // expose addModule for management
        methods.add(new String("addModule"));

        // set default values
        // componentProperties.putAll(SoapProperties.getDefaultProperties());
    }

    /**
     * Add a module to the Axis2 engine
     *
     * @param url
     * @return
     * @throws JBIException
     */
    public String addModule(String url) throws Exception {
        return bootstrapOperations.addModule(url);
    }
}
