/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap.listener.outgoing;

import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.stream.FactoryConfigurationError;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMText;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.client.Options;
import org.apache.axis2.client.ServiceClient;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.axis2.transport.http.HttpTransportProperties;
import org.apache.axis2.util.XMLUtils;
import org.objectweb.petals.binding.soap.SoapComponentContext;
import org.objectweb.petals.binding.soap.listener.SUProperties;
import org.objectweb.petals.component.framework.AbstractComponent;
import org.objectweb.petals.component.framework.PEtALSCFException;
import org.objectweb.petals.component.framework.mbean.PerformanceNotification;
import org.objectweb.petals.component.framework.util.Exchange;
import org.objectweb.petals.component.framework.util.Extensions;
import org.objectweb.petals.component.framework.util.SourceHelper;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * An external web service dispatcher. This dispatcher send the JBI message to
 * an external web service. The service URL is specified in the PEtALS
 * extensions.
 *
 * @author chamerling - eBM WebSourcing
 *
 */
public class WSCaller extends AbstractJBIDispatcher {

    /**
     * The JBI component instance
     */
    private AbstractComponent component = null;

    /**
     * Creates a new instance of {@link WSCaller}
     *
     * @param soapContext
     * @param logger
     */
    public WSCaller(SoapComponentContext soapContext, Logger logger,
            AbstractComponent component) {
        super(soapContext, logger);
        this.component = component;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.binding.soap.listener.outgoing.JBIListenerDispatcher#process(org.objectweb.petals.component.common.util.MessageExchangeWrapper,
     *      org.objectweb.petals.common.descriptor.Extensions)
     */
    public void process(Exchange exchange,
            Extensions cdkExtensions) {
        String address = cdkExtensions.getPetalsExtensions().getProperty(
                SUProperties.ADDRESS);

        logger.fine("Calling external web service : " + address);

        // Get the incoming Normalized message
        NormalizedMessage in = exchange.getInMessage();
        if (in == null) {
            exchange.setError(new JBIException(
                    "Message exchange must handle an In normalized message"));
        } else {

            // Get the outgoing Normalized message
            NormalizedMessage out = exchange.getOutMessage();

            // Get operation
            QName jbiOperation = exchange.getOperation();
            QName soapAction = getSOAPAction(jbiOperation, cdkExtensions);

            // create Options for the stub
            Options options = createOptions(address, soapAction, cdkExtensions);

            // create service client used to invoke WS
            ServiceClient serviceClient = null;
            try {
                serviceClient = Axis2Utils.createServiceClient(exchange
                        .getExchangePattern(), soapAction);

                serviceClient.setOptions(options);

                // create the in body
                OMElement inBodyElement = createSOAPMessage(in, exchange
                        .getOperation(), serviceClient, cdkExtensions);

                // engage the Axis2 modules
                engageModules(address, serviceClient);

                /* Send the message */

                if (exchange.isInOnlyPattern()) {
                    // send as InOnly message
                    serviceClient.fireAndForget(soapAction, inBodyElement);

                } else if (exchange.isRobustInOnlyPattern()) {
                    // send as RobustInOnly message
                    // If the robust in only exchange throws a Fault
                    // It will be catch by the component-common framework
                    serviceClient.sendRobust(soapAction, inBodyElement);

                } else if (exchange.isInOptionalOutPattern()
                        || exchange.isInOutPattern()) {
                    // Performance notification creation
                    PerformanceNotification performanceNotification = null;
                    performanceNotification = this.component
                            .createPerformanceNotificationSendingToOutside(
                                    address, exchange);

                    // send as In(optional)Out message
                    OMElement outBodyElement;
                    try {
                        outBodyElement = serviceClient.sendReceive(
                                soapAction, inBodyElement);
                    } catch (AxisFault e) {
                        throw e;
                    } finally {
                        // Performance notification sending
                        this.component
                                .sendPerformanceNotification(performanceNotification);
                    }

                    // if msg exchange required a response, set it
                    if (outBodyElement != null
                            && outBodyElement.toString().length() > 0) {
                        
                        Source responseSource = SourceHelper
                                .createSource(XMLUtils.toDOM(outBodyElement)
                                        .getOwnerDocument());

                        // if (out == null) {
                        // out = exchange.createMessage();
                        // exchange.setOutMessage(out);
                        //                        }
                        out.setContent(responseSource);

                        // create the JBI response from the WS one
                        marshaller
                                .copySOAPAttachementTOJBI(outBodyElement, out);
                    }
                } else {
                    throw new MessagingException(
                            "JBIListener - MEP not recognized : "
                                    + exchange.getPattern().toString());
                }
            } catch (Exception e) {
                try {
                    exchange.setFault(e);
                } catch (MessagingException e1) {
                    logger.log(Level.SEVERE, "Can't return fault to consumer",
                            e1);
                }
            }
        }
    }

    /**
     * Create the SOAP options of the outgoing SOAP message
     *
     * @param address
     * @param operation
     * @param extensions
     * @return
     */
    protected Options createOptions(String address, QName soapAction,
            Extensions extensions) {

        Options options = new Options();

        // set destination address
        options.setTo(new EndpointReference(address));

        // set the soap action
        if (soapAction != null) {
            options.setAction(soapAction.toString());
        }

        // set the timeout
        long timeout = SUProperties.retrieveTimeOut(extensions
                .getPetalsExtensions(), -1L);
        if (timeout != -1L) {
            options.setTimeOutInMilliSeconds(timeout);
        }

        // Get the soapEnvelopeNamespaceURI version to use - optional, default
        // is 1.1 version
        String soapEnvelopeNamespaceURI = retrieveSOAPEnvelopeNamespaceURI(extensions);
        options.setSoapVersionURI(soapEnvelopeNamespaceURI);

        // set transport
        options.setTransportInProtocol(Constants.TRANSPORT_HTTP);

        // get proxy settings if they are defined in the extensions
        HttpTransportProperties.ProxyProperties proxyProperties = retrieveProxySettings(extensions);
        if (proxyProperties != null) {
            options.setProperty(HTTPConstants.PROXY, proxyProperties);
        }

        // TODO set From address
        // options.setFrom(new EndpointReference(fromAddress));

        // FIXME, seems to be unusefull
        // options.setProperty(
        // org.apache.axis2.transport.http.HTTPConstants.CHUNKED,
        // org.apache.axis2.Constants.VALUE_FALSE);
        return options;
    }

    /**
     * Create an Axis OMElement representing the given body. The JBI content is
     * used as root element. All the JBI attachments are added as root element
     * children.
     *
     * @param body
     * @return
     * @throws FactoryConfigurationError
     * @throws XMLStreamException
     */
    @SuppressWarnings("unchecked")
    protected OMElement createSOAPMessage(NormalizedMessage nm,
            QName operation, ServiceClient client, Extensions extensions)
            throws XMLStreamException, FactoryConfigurationError,
            PEtALSCFException {

        // create the root element
        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMElement document = buildSOAPBody(nm.getContent(), operation,
                extensions);

        // add attachments if any
        if (nm.getAttachmentNames().size() > 0) {
            // enable MTOM
            client.getOptions().setProperty(
                    Constants.Configuration.ENABLE_MTOM, Constants.VALUE_TRUE);

            // Add JBI attachments to the document element
            Set<String> names = nm.getAttachmentNames();
            for (String key : names) {
                DataHandler attachment = nm.getAttachment(key);
                OMText attach = fac.createOMText(attachment, true);
                document.addChild(attach);
            }
        }
        return document;
    }

    /**
     * TODO Enable the client to contact a service over an SSL connection. The
     * required values are in the PEtALS extensions if the SU has been
     * parametrized to enable SSL.
     *
     * @param extensions
     */
    protected void enableSSLSupport(Extensions extensions) {
        String trustStore = extensions.getPetalsExtensions().getProperty(
                "ssl-trustStore");
        String trustStorePassword = extensions.getPetalsExtensions()
                .getProperty("ssl-trustStorePassword");
        String keyStore = extensions.getPetalsExtensions().getProperty(
                "ssl-keyStore");
        String keyStorePassword = extensions.getPetalsExtensions().getProperty(
                "ssl-keyStorePassword");

        if (trustStore != null && trustStorePassword != null
                && keyStore != null && keyStorePassword != null) {
            // TODO
        }
    }

    /**
     * Get the operation for the output SOAP message. The policy is : If the JBI
     * contains an operation, use it, else get the SOAP action from the
     * extensions, else SOAP action is empty.
     *
     * @param jbiOperation
     * @param extensions
     * @return
     */
    protected QName getSOAPAction(QName jbiOperation, Extensions extensions) {
        QName soapAction = null;

        if (jbiOperation != null) {
            soapAction = jbiOperation;
        } else {
            soapAction = retrieveDefaultSOAPAction(extensions);
        }
        return soapAction;
    }

    /**
     * Build the SOAP body. The SOAP operation is added as root element if it is
     * configured in the petals extensions.
     *
     * @param body
     * @param operation
     * @param extensions
     * @return
     */
    protected OMElement buildSOAPBody(Source body, QName operation,
            Extensions extensions) throws PEtALSCFException,
            XMLStreamException {
        Document contentSrc = SourceHelper.createDocument(body);
        Document contentDst = contentSrc;

        boolean addRoot = retrieveAddRoot(extensions);

        if (addRoot) {
            // add the operation to the payload
            if (operation != null) {
                String prefix = operation.getPrefix();
                if (prefix == null || prefix.length() == 0) {
                    prefix = "soapbc";
                }
                
                try {
                    final DocumentBuilderFactory documentFactory = DocumentBuilderFactory
                            .newInstance();
                    documentFactory.setNamespaceAware(true);

                    final DocumentBuilder documentBuilder = documentFactory
                            .newDocumentBuilder();
                    contentDst = documentBuilder.newDocument();

                    final Element operationElement = contentDst.createElementNS(
                            operation.getNamespaceURI(), operation.getLocalPart());
                    operationElement.insertBefore(contentSrc, null);
                    
                    contentDst.appendChild(operationElement);
                } catch (DOMException e) {
                    throw new PEtALSCFException(e);
                } catch (ParserConfigurationException e) {
                    throw new PEtALSCFException(e);
                }
                
            }
        }

        // create OMElement from content
        XMLStreamReader parser = XMLInputFactory.newInstance()
                .createXMLStreamReader(new DOMSource(contentDst));
        StAXOMBuilder builder = new StAXOMBuilder(parser);

        OMElement document = builder.getDocumentElement();

        return document;
    }

    /**
     * Engage the modules for the service client. The modules are available on
     * the service unit listener since they have been defined during SU
     * deployment.
     *
     * @param serviceClient
     * @throws AxisFault
     */
    private void engageModules(String address, ServiceClient serviceClient)
            throws AxisFault {
        List<String> modules = soapContext.getModules(address);
        if (modules == null) {
            return;
        }

        for (String name : modules) {
            logger.info("Engaging module " + name);
            serviceClient.engageModule(name);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.binding.soap.listener.outgoing.JBIListenerDispatcher#getDispatcherType()
     */
    public String getDispatcherType() {
        return DEFAULT_DISPATCHER;
    }

}
